module NRiF
! from: Numerical Recipies (in Fortran), Press, WH et al.

implicit none

public :: SPLINT
public :: SPLIN2    
public :: SPLIE2 
public :: SPLINE 
public :: HUNT
public :: LOCATE

contains


!-------------------------------------------------------------
! source: http://www.hackchina.com/en/r/47061/SPLINE.FOR__html
! variable definitions have subsequently been included.
! changes have been made so that U is dynamically allocated and large arrays can be handled
      SUBROUTINE SPLINE(X,Y,N,YP1,YPN,Y2)
      real*8    :: X,Y,YP1,YPN,Y2,SIG,P,QN,UN
      real*8, dimension(:), allocatable   ::  U
      integer   :: N,NMAX,I,K
      !PARAMETER (NMAX=100)  
      DIMENSION X(N),Y(N),Y2(N) !,U(NMAX)
      allocate(U(N))
      IF (YP1.GT..99E30) THEN
        Y2(1)=0.
        U(1)=0.
      ELSE
        Y2(1)=-0.5
        U(1)=(3./(X(2)-X(1)))*((Y(2)-Y(1))/(X(2)-X(1))-YP1)
      ENDIF
      DO 11 I=2,N-1
        SIG=(X(I)-X(I-1))/(X(I+1)-X(I-1))
        P=SIG*Y2(I-1)+2.
        Y2(I)=(SIG-1.)/P
        U(I)=(6.*((Y(I+1)-Y(I))/(X(I+1)-X(I))-(Y(I)-Y(I-1))/(X(I)-X(I-1)))/(X(I+1)-X(I-1))-SIG*U(I-1))/P
11    CONTINUE
      IF (YPN.GT..99E30) THEN
        QN=0.
        UN=0.
      ELSE
        QN=0.5
        UN=(3./(X(N)-X(N-1)))*(YPN-(Y(N)-Y(N-1))/(X(N)-X(N-1)))
      ENDIF
      Y2(N)=(UN-QN*U(N-1))/(QN*Y2(N-1)+1.)
      DO 12 K=N-1,1,-1
        Y2(K)=Y2(K)*Y2(K+1)+U(K)
12    CONTINUE
      RETURN
      deallocate(U)
      END SUBROUTINE

!-------------------------------------------------------------
! source: http://www.hackchina.com/en/r/47061/SPLINT.FOR__html
! variable definitions have subsequently been included
      SUBROUTINE SPLINT(XA,YA,Y2A,N,X,Y)
      integer   :: N, KLO, KHI, K
      real*8    :: X, Y, H, A, B
      real*8    :: XA, YA, Y2A
      DIMENSION XA(N),YA(N),Y2A(N)
      KLO=1
      KHI=N
1     IF (KHI-KLO.GT.1) THEN
        K=(KHI+KLO)/2
        IF(XA(K).GT.X)THEN
          KHI=K
        ELSE
          KLO=K
        ENDIF
      GOTO 1
      ENDIF
      H=XA(KHI)-XA(KLO)
      IF (H.EQ.0.) PAUSE 'Bad XA input.'
      A=(XA(KHI)-X)/H
      B=(X-XA(KLO))/H
      Y=A*YA(KLO)+B*YA(KHI)+((A**3-A)*Y2A(KLO)+(B**3-B)*Y2A(KHI))*(H**2)/6.
      RETURN
      END SUBROUTINE
!-------------------------------------------------------------
! source: http://www.hackchina.com/en/r/47061/SPLIN2.FOR__html
! variable definitions have subsequently been included
      SUBROUTINE SPLIN2(X1A,X2A,YA,Y2A,M,N,X1,X2,Y)
      integer   :: NN,M,N,J,K
      real*8    :: X1A,X2A,YA,Y2A,X1,X2,Y
      real*8    :: YTMP,Y2TMP,YYTMP
      real*8    :: dum1=1.0e30
      PARAMETER (NN=100) 
      DIMENSION X1A(M),X2A(N),YA(M,N),Y2A(M,N),YTMP(NN),Y2TMP(NN),YYTMP(NN) 
      DO 12 J=1,M
        DO 11 K=1,N
          YTMP(K)=YA(J,K)
          Y2TMP(K)=Y2A(J,K)
11      CONTINUE
        CALL SPLINT(X2A,YTMP,Y2TMP,N,X2,YYTMP(J))
12    CONTINUE
      CALL SPLINE(X1A,YYTMP,M,dum1,dum1,Y2TMP)
      CALL SPLINT(X1A,YYTMP,Y2TMP,M,X1,Y)
      RETURN
      END SUBROUTINE

!-------------------------------------------------------------
! like SPLIN2 but handles large arrays (LA)
! this is done by only evaluating at the necessary M indices. 
! I.e. a limited column is constructed (only 2 values).
      SUBROUTINE SPLIN2LA(X1A,X2A,YA,Y2A,M,N,X1,X2,Y)
      integer   :: NN,M,N,J,K,I
      integer   :: M0, M1
      real*8    :: X1A,X2A,YA,Y2A,X1,X2,Y
      real*8    :: YTMP,Y2TMP,YYTMP, X1TMP
      real*8    :: dum1=1.0e30
      PARAMETER (NN=100) 
      DIMENSION X1A(M),X2A(N),YA(M,N),Y2A(M,N),YTMP(NN),Y2TMP(NN),YYTMP(2) 
      DIMENSION X1TMP(2)

      ! determine start and end indices to use - i.e. indices limiting X1 - i.e. determine which rows are relevant
      M0 = 1
      CALL HUNT(X1A,M,X1,M0)
      M1 = M0 + 1

      I=0
      DO 12 J=M0,M1
        I=I+1
        X1TMP(I)=X1A(J)
        DO 11 K=1,N
          YTMP(K)=YA(J,K)
          Y2TMP(K)=Y2A(J,K)
        11 CONTINUE
        CALL SPLINT(X2A,YTMP,Y2TMP,N,X2,YYTMP(I))
      12 CONTINUE
      CALL SPLINE(X1TMP,YYTMP,2,dum1,dum1,Y2TMP)
      CALL SPLINT(X1TMP,YYTMP,Y2TMP,M1-M0+1,X1,Y)
      RETURN
      END SUBROUTINE
!-------------------------------------------------------------
! http://www.hackchina.com/en/r/47061/HUNT.FOR__html
! variable definitions have subsequently been included
      SUBROUTINE HUNT(XX,N,X,JLO)
      integer   :: N, JLO, JHI, INC, JM
      real*8    :: XX
      real*8    :: X
      DIMENSION XX(N)
      LOGICAL ASCND
      ASCND=XX(N).GT.XX(1)
      IF(JLO.LE.0.OR.JLO.GT.N)THEN
        JLO=0
        JHI=N+1
        GO TO 3
      ENDIF
      INC=1
      IF(X.GE.XX(JLO).EQV.ASCND)THEN
1       JHI=JLO+INC
        IF(JHI.GT.N)THEN
          JHI=N+1
        ELSE IF(X.GE.XX(JHI).EQV.ASCND)THEN
          JLO=JHI
          INC=INC+INC
          GO TO 1
        ENDIF
      ELSE
        JHI=JLO
2       JLO=JHI-INC
        IF(JLO.LT.1)THEN
          JLO=0
        ELSE IF(X.LT.XX(JLO).EQV.ASCND)THEN
          JHI=JLO
          INC=INC+INC
          GO TO 2
        ENDIF
      ENDIF
3     IF(JHI-JLO.EQ.1)RETURN
      JM=(JHI+JLO)/2
      IF(X.GT.XX(JM).EQV.ASCND)THEN
        JLO=JM
      ELSE
        JHI=JM
      ENDIF
      GO TO 3
      END SUBROUTINE HUNT
!-------------------------------------------------------------
! source: http://www.hackchina.com/en/r/47061/SPLIE2.FOR__html
! variable definitions have subsequently been included
      SUBROUTINE SPLIE2(X1A,X2A,YA,M,N,Y2A)
      integer   :: NN,M,N,J,K
      real*8    :: X1A,X2A,YA,Y2A
      real*8    :: YTMP, Y2TMP ! ORIGINAL
      real*8    :: dum1=1.0e30
      PARAMETER (NN=100) !ORIGINAL
      DIMENSION X1A(M),X2A(N),YA(M,N),Y2A(M,N),YTMP(NN),Y2TMP(NN) !ORIGINAL
!      DIMENSION X1A(M),X2A(N),YA(M,N),Y2A(M,N),YTMP(N),Y2TMP(N) !NEW
      DO 13 J=1,M
        DO 11 K=1,N
          YTMP(K)=YA(J,K)
11      CONTINUE
        CALL SPLINE(X2A,YTMP,N,dum1,dum1,Y2TMP)
        DO 12 K=1,N
          Y2A(J,K)=Y2TMP(K)
12      CONTINUE
13    CONTINUE
      RETURN
      END SUBROUTINE
!-------------------------------------------------------------
! source: http://www.hackchina.com/en/r/47061/LOCATE.FOR__html
!      SUBROUTINE LOCATE(XX,N,X,J)
!      integer   :: N, J, JL, JU, JM
!      real*8    :: XX, X
!      DIMENSION XX(N)
!      JL=0
!      JU=N+1
!10    IF(JU-JL.GT.1)THEN
!        JM=(JU+JL)/2
!        IF((XX(N).GT.XX(1)).EQV.(X.GT.XX(JM)))THEN
!          JL=JM
!        ELSE
!          JU=JM
!        ENDIF
!      GO TO 10
!      ENDIF
!      J=JL
!      RETURN
!      END SUBROUTINE
!-------------------------------------------------------------
! source: Numerical Recipies in Fortran
! http://apps.nrbook.com/fortran/index.html
      SUBROUTINE locate(xx,n,x,j)
      integer   :: n, j
      real*8    :: xx(n), x
      ! Given an array xx(1:n), and given a value x, returns a value j
      ! such that x is between xx(j) and xx(j+1). xx(1:n) must be monotonic, 
      ! either increasing or decreasing. j=0 or j=n is returned to indicate that 
      ! x is out of range
      integer :: jl, jm, ju
      jl=0
      ju=n+1
10    if(ju-jl.gt.1)THEN
        jm=(ju+jl)/2
        IF((xx(n).ge.xx(1)).eqv.(x.ge.xx(jm)))THEN
          jl=jm
        ELSE
          ju=jm
        ENDIF
      GO TO 10
      ENDIF
      if(x.eq.xx(1))then
        j=1
      else if(x.eq.xx(n))then
        j=n-1
      else
        j=jl
      endif
      RETURN
      END SUBROUTINE
!-------------------------------------------------------------
      integer function locate_checkbounds(n,j)
      integer :: n, j
      locate_checkbounds = 0
      if ((j==0).or.(j==n)) then
       ! results from locate are out of bounds. I.e. error
       locate_checkbounds = 1
      end if
      end function locate_checkbounds
!-------------------------------------------------------------
      SUBROUTINE LOCATE2ROW(XX,N,X,J,M,R)
! like LOCATE but XX is assumed to be 2D array (MxN). Search is carried out in row R
      integer   :: N, J, JL, JU, JM
      integer   :: M, R
      real*8    :: X
      real*8, dimension(M,N) :: XX
      JL=0
      JU=N+1
10    IF(JU-JL.GT.1)THEN
        JM=(JU+JL)/2
        IF((XX(R,N).GT.XX(R,1)).EQV.(X.GT.XX(R,JM)))THEN
          JL=JM
        ELSE
          JU=JM
        ENDIF
      GO TO 10
      ENDIF
      J=JL
      RETURN
      END SUBROUTINE
!-------------------------------------------------------------



end module NRiF