Module genout_tools
use genout_typesnvars

implicit none

!--------------------------------------------------------------------
!--------------------------------------------------------------------
! Public/private declarations

public :: readline
public :: getwords
public :: getwords2
public :: get_parameters
public :: get_label
public :: inc_output_sensor
public :: decr_output_sensor


!DEC$ IF .NOT. DEFINED(WKIN)
public :: put_output
public :: inc_action_sensor
public :: decr_action_sensor
!DEC$ ENDIF

private :: lowercase
private :: inc_output_at
private :: inc_outputs

interface inc_type_array
  module procedure inc_output_at
  module procedure inc_outputs
  !DEC$ IF .NOT. DEFINED(WKIN)
  module procedure inc_dll_info_all
  !DEC$ ENDIF
end interface inc_type_array
!--------------------------------------------------------------------
!--------------------------------------------------------------------
! Type definitions

!--------------------------------------------------------------------

contains

!--------------------------------------------------------------------
!--------------------------------------------------------------------

subroutine inc_output_at(output_at)
! subroutine that increases the size of output_at%out array by one and keeps the old values
  Type (Toutput_at), intent(inout) :: output_at

  Type (Toutput_at) :: temp
  integer           :: i

  temp%nout=output_at%nout+1
  allocate(temp%out(temp%nout))
  if (output_at%nout>0) then
    do i=1,output_at%nout
      temp%out(i)=output_at%out(i)
    enddo
	deallocate(output_at%out)
	output_at%nout=temp%nout
	allocate(output_at%out(output_at%nout))
	output_at=temp
  else
	output_at%nout=1
	allocate(output_at%out(output_at%nout))
  endif
!  deallocate(temp%out)
end subroutine inc_output_at

!--------------------------------------------------------------------

subroutine inc_outputs(outputs)
! subroutine that increases the size of outputs%output array by one and keeps the old values
  Type (Toutputs), intent(inout) :: outputs

  Type (Toutputs)   :: temp
  integer           :: i

  temp%nout=outputs%nout+1
  allocate(temp%output(temp%nout))
  if (outputs%nout>0) then
    do i=1,outputs%nout
      temp%output(i)=outputs%output(i)
    enddo
	deallocate(outputs%output)
	outputs%nout=temp%nout
	allocate(outputs%output(outputs%nout))
	outputs=temp
  else
	outputs%nout=1
	allocate(outputs%output(outputs%nout))
  endif
end subroutine inc_outputs

!--------------------------------------------------------------------

subroutine inc_output_sensor(output,size)
! subroutine that increases the size of outputs%output array by size and keeps the old values
  Type (Toutput), intent(inout) :: output
  integer,intent(in)			:: size

  Type (Toutput)   :: temp
  integer           :: i

  temp%nsensors=output%nsensors+size
  allocate(temp%sensor(temp%nsensors))
  if (output%nsensors>0) then
    do i=1,output%nsensors
      temp%sensor(i)=output%sensor(i)
    enddo
	deallocate(output%sensor)
	output%nsensors=temp%nsensors
	allocate(output%sensor(output%nsensors))
	output%sensor=temp%sensor
  else
	output%nsensors=size
	allocate(output%sensor(output%nsensors))
  endif
end subroutine inc_output_sensor

!--------------------------------------------------------------------

subroutine decr_output_sensor(output,size)
! subroutine that decreases the size of outputs%output array by size and keeps the old values
  Type (Toutput), intent(inout) :: output
  integer,intent(in)			:: size

  Type (Toutput)   :: temp
  integer           :: i

  
  if (output%nsensors>1) then
    temp%nsensors=output%nsensors-size
    allocate(temp%sensor(temp%nsensors))
    do i=1,temp%nsensors
      temp%sensor(i)=output%sensor(i)
    enddo
	deallocate(output%sensor)
	output%nsensors=temp%nsensors
	allocate(output%sensor(output%nsensors))
	output%sensor=temp%sensor
  else
	deallocate(output%sensor)
	output%nsensors=0
  endif
end subroutine decr_output_sensor

!--------------------------------------------------------------------

!DEC$ IF .NOT. DEFINED(WKIN)

subroutine inc_dll_info_all(dll_info_all)
use dll_types
! subroutine that increases the size of dll_info%output and dll_info%input array by one and keeps the old values
  Type (Tdll_info_all), intent(inout) :: dll_info_all

  Type (Tdll_info_all) :: temp
  integer          :: i

  temp%ndll=dll_info_all%ndll+1
  allocate(temp%dll_info(temp%ndll))
  if (dll_info_all%ndll>0) then
    do i=1,dll_info_all%ndll
      temp%dll_info(i)=dll_info_all%dll_info(i)
    enddo
	deallocate(dll_info_all%dll_info)
	dll_info_all%ndll=temp%ndll
	allocate(dll_info_all%dll_info(dll_info_all%ndll))
	allocate(temp%dll_info(dll_info_all%ndll)%output)
	allocate(temp%dll_info(dll_info_all%ndll)%input)
	dll_info_all=temp
  else
	dll_info_all%ndll=1
	allocate(dll_info_all%dll_info(dll_info_all%ndll))
	allocate(dll_info_all%dll_info(dll_info_all%ndll)%output)
	allocate(dll_info_all%dll_info(dll_info_all%ndll)%input)
  endif
end subroutine inc_dll_info_all

!DEC$ ENDIF

!--------------------------------------------------------------------
!--------------------------------------------------------------------

!DEC$ IF .NOT. DEFINED(WKIN)

subroutine inc_action_sensor(action,size)
  use dll_types
! subroutine that increases the size of actions%sensor array by size and keeps the old values
  Type (Taction), intent(inout) :: action
  integer,intent(in)			:: size

  Type (Taction)   :: temp
  integer           :: i

  temp%nsensors=action%nsensors+size
  allocate(temp%sensor(temp%nsensors))
  if (action%nsensors>0) then
    do i=1,action%nsensors
      temp%sensor(i)=action%sensor(i)
    enddo
	deallocate(action%sensor)
	action%nsensors=temp%nsensors
	allocate(action%sensor(action%nsensors))
	action%sensor=temp%sensor
  else
	action%nsensors=size
	allocate(action%sensor(action%nsensors))
  endif
end subroutine inc_action_sensor

!DEC$ ENDIF

!--------------------------------------------------------------------
!--------------------------------------------------------------------

!DEC$ IF .NOT. DEFINED(WKIN)

subroutine decr_action_sensor(action,size)
  use dll_types
! subroutine that decreases the size of action%sensor array by size and keeps the old values
  Type (Taction), intent(inout) :: action
  integer,intent(in)			:: size

  Type (Taction)   :: temp
  integer           :: i

  
  if (action%nsensors>1) then
    temp%nsensors=action%nsensors-size
    allocate(temp%sensor(temp%nsensors))
    do i=1,temp%nsensors
      temp%sensor(i)=action%sensor(i)
    enddo
	deallocate(action%sensor)
	action%nsensors=temp%nsensors
	allocate(action%sensor(action%nsensors))
	action%sensor=temp%sensor
  else
	deallocate(action%sensor)
	action%nsensors=0
  endif
end subroutine decr_action_sensor

!DEC$ ENDIF

!--------------------------------------------------------------------
!--------------------------------------------------------------------
!
subroutine readline(masterfile,line,init)
Type (Tmasterfile),pointer    :: masterfile
Type (Tline),intent(inout)    :: line
logical,intent(inout)         :: init


if (init==.true.) then
  masterfile%linenr=0
  init=.false.
endif

read(unit=masterfile%unit,fmt='(A)',end=99) line
masterfile%linenr=masterfile%linenr+1

call lowercase(line)

return
99 continue
write(0,*) '*** INFO *** End of file reached. Line number',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
line%line='exit ;'
end subroutine readline

!--------------------------------------------------------------------
!--------------------------------------------------------------------

subroutine getwords(line,words,masterfile)
use dfport
use genout_tools2, only : getfreeunit
! divides a line into several words spaced by zero, line ended with ;
character*(*),intent(inout)     :: line
Type (Tword), intent(inout)     :: words
Type (Tmasterfile),pointer      :: masterfile

character :: linestop,blank,tab
parameter (linestop=';')
parameter (blank   =' ')
parameter (tab     ='	')

integer :: first, &
           last, &
		   endp, &
		   blankpos, &
		   tabpos

integer :: tmp_blank,tmp_tab

logical :: out

integer :: error, info_array(12)
logical :: init,cont
Type (Tline) :: line2

cont=.false.
DO while (cont==.false.)
cont=.true.

init=.false.

words%word_vec(:)=''

line=adjustl(line)

endp=index(line,linestop)
first=1
words%nwords=0
out=.false.

if (endp>1) then
  do while (out==.false.)
    blankpos=index(line(first:endp),blank)
	tabpos=index(line(first:endp),tab)
	if (tabpos/=0) then
	  if (blankpos/=0) then
	    blankpos=min(blankpos,tabpos)
	  else 
	    blankpos=tabpos
	  endif
	endif
!   If the restline is without blanks, the restline is stored and
!   the division in words is stopped ...
    if (blankpos==0) then
      words%nwords=words%nwords+1
      words%word_vec(words%nwords)=line(first:endp-1)
	  out=.true.
      exit
    endif

!   If a blank is found a word is stored ...
    if (blankpos>1) then
      words%nwords=words%nwords+1
      last=first+blankpos-2
      words%word_vec(words%nwords)=line(first:last)
      first=first+blankpos
    endif

!   If search has reached and end of line or if too many words have
!   been found then stop
    if (first>=endp) then
	  out=.true.
	  exit
	elseif (words%nwords>size(words%word_vec)) then
      write(0,*) 'More than ',size(words%word_vec),' words in command. Exceeding words ignored. Line ',masterfile%linenr
	  exit
	  out=.true.
    endif 

!   else remove blanks in restline by incrementing the first position
    tmp_blank=index(line(first:endp),blank)
	tmp_tab=index(line(first:endp),tab)
	if (tmp_tab/=0) then
	  if (tmp_blank/=0) then
	    tmp_blank=min(tmp_blank,tmp_tab)
	  else 
	    tmp_blank=tmp_tab
	  endif
	endif

!	if (tmp_tab/=0) tmp_blank=min(tmp_blank,tmp_tab)
    do while (tmp_blank==1)
      first=first+1
	  if (first>=endp) then
	    out=.true.
  	    exit
	  endif
      tmp_blank=index(line(first:endp),blank)
	  tmp_tab=index(line(first:endp),tab)
	  if (tmp_tab/=0) tmp_blank=min(tmp_blank,tmp_tab)
    enddo
  enddo
elseif (endp==0) then 
  write(0,*) '*** ERROR *** No line termination in command line ',masterfile%linenr
endif

if (trim(words%word_vec(1))=='continue_in_file') then
  write(0,*) 'Main input file is continued in file :',trim(words%word_vec(2))
  if (.not.associated(masterfile%child)) allocate(masterfile%child)
  masterfile%child%mother=>masterfile
  masterfile%child%name=trim(words%word_vec(2))
  masterfile%child%level=masterfile%level+1
  masterfile=>masterfile%child

  call getfreeunit(masterfile%unit)           
  error=stat(masterfile%name,info_array)
  if (error/=0) then
    GOTO 10
  else 
    open (unit=masterfile%unit, file=masterfile%name,err=10)  ! commandfile is opened
  endif
  call readline(masterfile,line2,init)
  line=line2%line
  cont=.false.
  cycle    ! repeat contents of getwords
elseif (trim(words%word_vec(1))=='exit') then
  if (masterfile%level.gt.0) then
    write(0,*) 'Main input file is returned to :',trim(masterfile%mother%name)
	close(masterfile%unit)
    masterfile=>masterfile%mother
	masterfile%child=>null()
    call readline(masterfile,line2,init)
    line=line2%line
	cont=.false.
    cycle   ! repeat contents of getwords
  endif
endif

ENDDO

return
  10 write(0,*) 'Error opening file: ',trim(masterfile%name)
  stop
end subroutine getwords

!--------------------------------------------------------------------
!--------------------------------------------------------------------

subroutine getwords2(line,words)
! As getwords, but no line end used
character*(*),intent(inout)     :: line
Type (Tword), intent(inout)     :: words

character :: blank,tab
parameter (blank   =' ')
parameter (tab     ='	')

integer :: first, &
           last, &
		   endp, &
		   blankpos, &
		   tabpos

integer :: tmp_blank,tmp_tab

logical :: out

words%word_vec(:)=''

line=adjustl(line)

endp=len(line)
first=1
words%nwords=0
out=.false.

if (endp>1) then
  do while (out==.false.)
    blankpos=index(line(first:endp),blank)
	tabpos=index(line(first:endp),tab)
	if (tabpos/=0) then
	  if (blankpos/=0) then
	    blankpos=min(blankpos,tabpos)
	  else 
	    blankpos=tabpos
	  endif
	endif
!   If the restline is without blanks, the restline is stored and
!   the division in words is stopped ...
    if (blankpos==0) then
      words%nwords=words%nwords+1
      words%word_vec(words%nwords)=line(first:endp-1)
	  out=.true.
      exit
    endif

!   If a blank is found a word is stored ...
    if (blankpos>1) then
      words%nwords=words%nwords+1
      last=first+blankpos-2
      words%word_vec(words%nwords)=line(first:last)
      first=first+blankpos
    endif

!   If search has reached and end of line or if too many words have
!   been found then stop
    if (first>=endp) then
	  out=.true.
	  exit
	elseif (words%nwords>size(words%word_vec)) then
      write(0,*) '*** ERROR *** More than ',size(words%word_vec),' words in command. Exceeding words ignored'
	  exit
	  out=.true.
    endif 

!   else remove blanks in restline by incrementing the first position
    tmp_blank=index(line(first:endp),blank)
	tmp_tab=index(line(first:endp),tab)
	if (tmp_tab/=0) then
	  if (tmp_blank/=0) then
	    tmp_blank=min(tmp_blank,tmp_tab)
	  else 
	    tmp_blank=tmp_tab
	  endif
	endif
!	if (tmp_tab/=0) tmp_blank=min(tmp_blank,tmp_tab)
    do while (tmp_blank==1)
      first=first+1
	  if (first>=endp) then
	    out=.true.
  	    exit
	  endif
      tmp_blank=index(line(first:endp),blank)
	  tmp_tab=index(line(first:endp),tab)
	  if (tmp_tab/=0) tmp_blank=min(tmp_blank,tmp_tab)
    enddo
  enddo
endif

end subroutine getwords2

!--------------------------------------------------------------------
!--------------------------------------------------------------------

subroutine get_parameters(words,nstart,npars,out_vec,err)
Type (Tword), intent(in)            :: words
integer , intent(in)                :: nstart, &
                                       npars
real*8, dimension(:), intent(inout) :: out_vec
logical, intent(inout)              :: err

integer :: i

if (words%nwords<(nstart+npars-1)) then
  write(0,*) '*** ERROR *** Not correct number of parameters in command line'
  err=.true.
  goto 99
else
  do i=nstart,nstart+npars-1
    read(words%word_vec(i),Fmt=*,err=99) out_vec(i-nstart+1)
    err=.false.
  enddo
endif
return
99 write(0,*) '*** ERROR *** Error reading parameters in command line: ',trim(words%word_vec(1)),' ',trim(words%word_vec(2))
   err=.true.
end subroutine get_parameters

!--------------------------------------------------------------------
subroutine get_label(words,label)
! This subroutine searches the input words for a label which should be enclosed in #'s
Type (Tword),intent(in)              :: words
character*(*),intent(out)            :: label

logical :: mark1, &
           mark2
integer   :: i

label=''
mark1=.false.
mark2=.false.
do i=1,size(words%word_vec)
  if (mark1==.false.) then
    if (trim(words%word_vec(i))=='#') then
      mark1=.true.
	endif
  else
    if (trim(words%word_vec(i))=='#') then
      mark2=.true.
	else
      label=trim(label)//' '//trim(words%word_vec(i))
	endif
  endif
  if (mark2==.true.) then
    cycle
  endif
enddo
end subroutine get_label
!--------------------------------------------------------------------

!--------------------------------------------------------------------
subroutine get_only_exclude(words,only_excl,err)
! This subroutine searches the input word only and uses the following number to accept for output, the rest are excluded
Type (Tword),intent(in)              :: words
logical,dimension(:)				 :: only_excl
logical,intent(inout)				 :: err

integer   :: i
integer	  :: number

do i=1,size(words%word_vec)
  if (trim(words%word_vec(i))=="only") then
    read(words%word_vec(i+1),Fmt=*,err=99) number
	if (number<1) goto 99
	if (number>size(only_excl)) goto 99
    only_excl(:)=.true.
    only_excl(number)=.false.
	exit
  elseif (trim(words%word_vec(i))==";") then
    exit
  endif
enddo

return
99 write(0,*) '*** ERROR *** Error reading "only" parameter in command line: ',trim(words%word_vec(1)),' ',trim(words%word_vec(2))
   err=.true.
end subroutine get_only_exclude
!--------------------------------------------------------------------
!--------------------------------------------------------------------

Function onoff(word,masterfile)
character*wordlenght, intent(in) :: word
Type (Tmasterfile), pointer      :: masterfile
character*3                      :: onoff

onoff=trim(word)
if (onoff=='on') then
  continue
elseif (onoff=='off') then
  continue
else
  write(0,*) '*** ERROR *** Error in regarding on/off parameter in line ',masterfile%linenr
endif

end function onoff
!--------------------------------------------------------------------
!--------------------------------------------------------------------

subroutine lowercase(line)
Type (Tline), intent(inout) :: line

integer   :: i

do i=1,len(line%line)
  if ((line%line(i:i)>='A').and.(line%line(i:i)<='Z')) then
    line%line(i:i)=char(ichar(line%line(i:i))+32)
  endif
enddo
end subroutine lowercase

!--------------------------------------------------------------------
!--------------------------------------------------------------------

!DEC$ IF .NOT. DEFINED(WKIN)

subroutine put_hawc_dll_output(nr,ov,status,output)
  integer, intent(in)            :: nr
  Type (Toutvar), intent(in)     :: ov
  character*(*),intent(in)       :: status
  Type (Toutput),intent(inout)   :: output

  if (nr>output%nov) goto 10

  if ((trim(status)=='').or.(trim(status)=='1')) then
    output%ov(nr)=ov
  endif

  return
  10 write(0,*) 'More sensors in output than defined for output format: ',output%data_format

end subroutine put_hawc_dll_output

!DEC$ ENDIF

!--------------------------------------------------------------------
!--------------------------------------------------------------------

!DEC$ IF .NOT. DEFINED(WKIN)

subroutine put_output(nr,ov,status,output)
use genout_hawc_file
use genout_hawcbin_file
integer, intent(in)         :: nr
Type (Toutvar), intent(in)  :: ov
character*(*),intent(in)    :: status
Type (Toutput),intent(inout):: output

  if (nr>output%nsensors) goto 10
  select case (output%data_format)
  !-------------------
  case ('hawc_ascii')
    call put_hawc_output(output%sensor(nr)%output_number,ov,status,output)
  case ('hawc_binary')
    call put_hawcbin_output(output%sensor(nr)%output_number,ov,status,output)
  !-------------------
    case ('hawc_dll')
      call put_hawc_dll_output(output%sensor(nr)%output_number,ov,status,output)
  case default
    write(0,*) '*** ERROR *** Unknown data format for output - ignoring'
  end select

return
10 write(0,*) 'More sensors in output than defined for output format: ',output%data_format

end subroutine put_output

!DEC$ ENDIF

!--------------------------------------------------------------------
!--------------------------------------------------------------------

subroutine update_output_numbers(output)
Type (Toutput),intent(inout)	:: output

integer							:: nr,i

nr=1
do i=1,output%nsensors
  if (.not.output%sensor(i)%only_exclude) then
    output%sensor(i)%output_number=nr
	nr=nr+1
  endif
enddo
output%nout=output%sensor(output%nsensors)%output_number

end subroutine update_output_numbers

end module genout_tools
