module ireg_airy
use global_typesnvars

implicit none

public :: ireg_airy_input
public :: ireg_airy_init
public :: ireg_airy_sea_elevation
public :: ireg_airy_get_kinematics

private :: dispersion
private :: dispersion_deriv
private :: jonswap

contains

!-------------------------------------------------------------

Subroutine ireg_airy_input(masterfile,ireg)
use genout_tools
!use global_vars
use global_typesnvars, only : wd

Type (Tmasterfile), pointer :: masterfile
Type (Tireg),pointer		:: ireg

Type (Tline) :: line
Type (Tword) :: words

real*8, dimension(maxwords) :: rea

logical :: init, &
           out, &
           err, &
           all_ok, &
		   stretching_control, &
		   spectrum_control, &
		   coef_control

init=.false.
out=.false.
stretching_control=.false.
spectrum_control=.false.
coef_control=.false.

do while (out==.false.)
  call readline(masterfile,line,init)        ! line is read
    call getwords(line%line,words,masterfile)                           ! words in line is retrieved
    select case (trim(words%word_vec(1)))
    
	case ('stretching')
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) then; goto 10; endif
      ireg%stretching=nint(rea(1))
      stretching_control=.true.

	case ('spectrum')
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) then; goto 10; endif
      ireg%spectrum=nint(rea(1))
      spectrum_control=.true.

	case ('jonswap')
      call get_parameters(words,2,3,rea,err)
      if (err==.true.) then; goto 10; endif
	  allocate(ireg%jonswap)
      ireg%jonswap%Hs=rea(1)
      ireg%jonswap%Tp=rea(2)
      ireg%jonswap%gamma=rea(3)

	case ('pm')
      call get_parameters(words,2,2,rea,err)
      if (err==.true.) then; goto 10; endif
	  allocate(ireg%pm)
      ireg%PM%Hs=rea(1)
      ireg%PM%Tz=rea(2)

	case ('wn')		! White noise
      call get_parameters(words,2,3,rea,err)
      if (err==.true.) then; goto 10; endif
	  allocate(ireg%wn)
      ireg%WN%var_target=rea(1)
      ireg%WN%f0=rea(2)
      ireg%WN%f1=rea(3)

	case ('coef')
      call get_parameters(words,2,2,rea,err)
      if (err==.true.) then; goto 10; endif
      ireg%N=nint(rea(1))
	  ireg%seednr=nint(rea(2))
      coef_control=.true.

	case ('spreading')
      call get_parameters(words,2,2,rea,err)
      if (err==.true.) then; goto 10; endif
	  ireg%spreading=nint(rea(1))
      ireg%s		=nint(rea(2))

	case ('pregen')
	  call get_parameters(words,2,1,rea,err)
      if (err==.true.) then; goto 10; endif
	  if (nint(rea(1)).eq.1) then
	    ireg%pregen=.true.
	    allocate(ireg%wfield)
		write(0,*) 'In Wkin_dll: Wave field will be precalculated'
	  else
	    ireg%pregen=.false.
	  endif
    
	case ('begin')
      select case (trim(words%word_vec(2)))

      case ('pregen_field')
	    if (.not.associated(ireg%wfield)) allocate(ireg%wfield)
        call get_wfield_input(masterfile,ireg%wfield)              ! input for BEMwake method
      case default
        write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
      end select

    case ('')
      continue

    case ('end')
      out=.true.
      ! Control of data given above is OK
      all_ok=.true.
      if ((stretching_control==.false.) .or. &
	      (spectrum_control==.false.)) then

        all_ok=.false.
      endif
	  if (ireg%spectrum.ne.3) then
	    if (coef_control==.false.) all_ok=.false.
      endif
      if (all_ok==.true.) then
        write(0,*) 'Wkin dll input commands read with succes'
      else
        write(0,*) '*** ERROR *** Not all needed wkin_dll ireg_airy input commands present - error'
      endif
    case (';')
      continue
    case default
      write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    end select
enddo
return
10  write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)

end subroutine ireg_airy_input

subroutine get_wfield_input(masterfile,wf)
use genout_tools
!use global_vars
Type (Tmasterfile), pointer :: masterfile
Type (Twavefield)          :: wf

Type (Tline) :: line
Type (Tword) :: words

real*8, dimension(maxwords) :: rea

logical :: init, &
           out, &
           err

init=.false.
out=.false.

do while (out==.false.)
  call readline(masterfile,line,init)        ! line is read
    call getwords(line%line,words,masterfile)                          ! words in line is retrieved
    select case (trim(words%word_vec(1)))

    case ('y_resolution')
      call get_parameters(words,2,2,rea,err)
      if (err==.true.) exit
      wf%ny=rea(1)
	  if (modulo(wf%ny,2).eq.0) then
	    write(0,*) 'In wkin_dll: Ny in y-resolution of wavefield has to be an uneven number. ',wf%ny+1,' is used instead of ',wf%ny
		wf%ny=wf%ny+1
	  endif
      wf%dy=rea(2)

    case ('t_resolution')
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
      wf%dt=rea(1)

    case ('z_resolution')
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
      wf%nz=rea(1)

    case ('x_range')
      call get_parameters(words,2,1,rea,err)
      if (err==.true.) exit
      wf%xrange=rea(1)

    case ('')
      continue

    case ('end')
      out=.true.
        write(0,*) 'In Wkin_dll, wfield commands read with succes'
    case default
      write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    end select
    if (err==.true.) then
      write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    endif
enddo

end subroutine get_wfield_input
!--------------------------------------------------------------------
!--------------------------------------------------------------------
!-------------------------------------------------------------

subroutine ireg_airy_init(ireg,t0,t1,dt)
use tools
use msimsl
Type (Tireg)	:: ireg
real*8			:: t0
real*8			:: t1
real*8			:: dt

integer			:: i,j
real*8			:: omega_min
real*8			:: omega_max
real*8			:: omega
real*8			:: Sp
real*8			:: sqr2
Type (Tcoef),pointer	:: c

integer			:: m
real*8			:: step
real*8,dimension(:),pointer	:: S
real*8			:: var_target,var_act,sum,elevation

real*8			:: K2s,value,value_old
integer         :: NR=1000   ! Number of parameters in order to evaluate cumulative function of spreading pdf
real*8,dimension(:),pointer :: FCUM ! cumulative function of spreading pdf
real*8,dimension(:),pointer :: angle ! angles corresponding to cumulative function

real*8,dimension(:),pointer :: randomvar
real*8,dimension(:),pointer :: elev_filtered=>null()
real*8, pointer       :: coef(:)

! Find fourier coefficients
CALL RNSET (ireg%seednr)
!if (associated(ireg%coef)) deallocate(ireg%coef)

allocate(ireg%coef(ireg%N))

if (ireg%spreading==0) then
  allocate(randomvar(ireg%N))
  CALL DRNUN (ireg%N, randomvar)
else
  allocate(randomvar(ireg%N*2))
  CALL DRNUN (ireg%N*2, randomvar)
endif

do i=1,ireg%N
  ireg%coef(i)%phi=randomvar(i)*2.0d0*pi-pi	! scale phases to be between -pi and pi
enddo

! Find min and max frequencies in spectra
omega_min=-1.0d0
omega_max=-1.0d0
omega=-1.0d-6

if (ireg%spectrum.eq.3) then  ! White noise
  omega_min=2.d0*pi*ireg%wn%f0
  omega_max=2.d0*pi*ireg%wn%f1
else
  do while (omega_max.le.0.0d0)
    omega=omega+0.02
    select case (ireg%spectrum)
    !--------------------------------------------------------
    case (1) ! Jonswap spectrum
      call jonswap(ireg%jonswap,omega,Sp)
      if ((Sp>1d-3).and.(omega_min.le.0.0d0)) then
        omega_min=omega
      elseif ((Sp<1.d-3).and.(omega>2*pi/ireg%jonswap%Tp)) then	! dette kriterie skal lige testes helt igennem
        omega_max=omega
      endif
    case (2) ! PM spectrum
      call PM(ireg%pm,omega,Sp)
      if ((Sp>1d-3).and.(omega_min.le.0.0d0)) then
        omega_min=omega
      elseif ((Sp<1.d-3).and.(omega>2*pi/ireg%pm%Tz)) then	! dette kriterie skal lige testes helt igennem
        omega_max=omega
      endif
    !--------------------------------------------------------
    case default
      continue
    end select
  enddo
endif

! Find omega interval
ireg%omega_min=omega_min
ireg%omega_step=(omega_max-omega_min)/(Ireg%N-1)

! Find Amplitudes
sqr2=dsqrt(2.0d0)
do i=1,ireg%N
  c=>ireg%coef(i)
  c%omega=ireg%omega_min+(i-1)*ireg%omega_step
  select case (ireg%spectrum)
  case (1) ! Jonswap spectrum
    call jonswap(ireg%jonswap,c%omega,Sp)
  case (2) ! Pm spectrum
    call pm(ireg%pm,c%omega,Sp)
  case (3) ! White noise spectrum
    Sp=ireg%wn%var_target/omega_max
  case default
  end select
  c%A=sqr2*dsqrt(Sp*ireg%omega_step)
enddo

! find wave numbers
do i=1,ireg%N
  c=>ireg%coef(i)
  c%k=newton_raphson(2.0d0,c%omega,wd%wdepth,dispersion,dispersion_deriv,1.0d-6)
enddo

!------------------------------------------------
! If directional spreading is requested
if (ireg%spreading.eq.1) then
  K2s=(2**(2*ireg%s-1)*FAC(ireg%s)*FAC(ireg%s-1)) / (pi*FAC(2*ireg%s-1))

  ! Find cumulative distribution of f(x)=K2s*(cos(x))^2s
  step=pi/(nr-1.d0)
  allocate(FCUM(nr))
  allocate(angle(nr))
  sum=0.d0
  value_old=0.d0
  do j=1,nr
    angle(j)=(j-1)*step-pi/2.d0
	value=K2s*(dcos(angle(j))**(2.0d0*ireg%s))
    sum=sum+(value+value_old)/2.d0*step
	value_old=value
	FCUM(j)=sum
  enddo

  ! Find directions
  do i=1,ireg%N
    c=>ireg%coef(i)
	c%dir=Getvalue(angle,Fcum,Nr,randomvar(i+ireg%N),2)
  enddo
endif

! Scale amplitudes to obtain correct variance
! Find target variance
m=nint(ireg%N*10/2.d0)					! 10 times higher resolution
omega_min=omega_min*0.75d0
if (ireg%spectrum.eq.3) then
  continue                              ! max frequency is absolute - not to be changed here
else
  omega_max=omega_max*1.25d0
endif

step=(omega_max-omega_min)/(2*m-1)		! 
allocate(S(0:2*m))
do i=0,2*m
  omega=omega_min+i*step
  select case (ireg%spectrum)
  case (1) ! Jonswap spectrum
    call jonswap(ireg%jonswap,omega,S(i))
  case (2) ! PM spectrum
    call pm(ireg%pm,omega,S(i))
  case (3) ! White noise spectrum
    S(i)=ireg%wn%var_target/omega_max
  case default
  end select
enddo
var_target=simpson_int(omega_min,omega_max,m,S(:))
if (ireg%spectrum.eq.3) var_target=ireg%wn%var_target

! Find actual variance
m=nint((t1-t0)/dt)
sum=0.d0
do j=1,m
  wd%time=t0+(j-1)*dt

  elevation=0
  do i=1,ireg%N
    c=>ireg%coef(i)
    elevation=elevation+c%A*dsin(c%omega*wd%time+c%phi)   ! here elevation is positive upwards
  enddo

  sum=sum+elevation**2
enddo
wd%time=0.d0
var_act=sum/(m-1)

! Scale coef
ireg%coef(:)%A=ireg%coef(:)%A*dsqrt(var_target/var_act)

! Deallocate variables
angle=>null()
FCUM=>null()
randomvar=>null()
S=>null()
elev_filtered=>null()
coef=>null()
!
! Pre-generate waves if requested
! 
if (ireg%pregen) then
  call ireg_airy_pregen(ireg,t0,t1,omega_max)
  ireg%coef=>null()
endif
!
end subroutine ireg_airy_init

!-----------------------------------
subroutine ireg_airy_pregen(ireg,t0,t1,omega_max)
!use tools
Type (Tireg)	:: ireg
real*8			:: t0
real*8			:: t1
real*8			:: omega_max
! Locals
type (Twavefield),pointer	:: wf=>null()
integer,dimension(1)		:: q
Type (Tcoef),pointer	:: c
real*8						:: eps=1.d-7
integer			:: i,y,z,t
real*8,dimension(3)			:: pos_h=0.d0
real*8						:: time
real*8						:: alfa
real*8						:: cosh1,sinh1,sinh2,cosh2
real*8						:: tmp1,tmp2,tmp3,tmp4,tmp5,tmp
!real*4,pointer				:: a_sin1=>null()
!real*4,pointer				:: a_cos1=>null()
real*8,pointer				:: a_sin1=>null()
real*8,pointer				:: a_cos1=>null()
!
! Calculate wave field and store in memory
if (ireg%pregen) then
  if (.not.associated(ireg%wfield)) allocate(ireg%wfield)
  Write(0,*) 'In Wkin_dll, calculating wave field'
  wf=>ireg%wfield

  ! find phase velocity of largest wave component
  q=MAXLOC(ireg%coef(:)%A)
  if (q(1).eq.1) q(1)=int(ireg%N/2)      ! inserted mainly for white noise waves where nothing really make any sence
  c=>ireg%coef(q(1))
  wf%groupvel=0.5d0*dsqrt(wd%g/c%k)*(c%k*wd%wdepth+dtanh(c%k*wd%wdepth)-c%k*wd%wdepth*dtanh(c%k*wd%wdepth)**2)/ &
              dsqrt(dtanh(c%k*wd%wdepth))

  wf%t0=wf%xrange/wf%groupvel
  ! Set dimensions for wave field
  if (wf%dt.eq.0.d0) then
    wf%dt=2*pi/omega_max/10.0d0   ! 10 points/wave period
  endif
  wf%nt=nint((t1-t0+2.d0*wf%t0)/wf%dt)

  allocate(wf%kdata(wf%nt,wf%nz,wf%ny))
  allocate(wf%surf(wf%nt,wf%ny))
  if (wf%ny.eq.1) wf%dy=eps

  ! Calculate elevation and temp alfa variables
  do y=1,wf%ny
    pos_h(2)=-(wf%ny-1)/2.d0*wf%dy+(y-1)*wf%dy
	pos_h(1)=0.d0
    do t=1,wf%nt
	  !time=(t-1)*wf%dt
	  time=(t-1)*wf%dt-wf%t0
	  allocate(wf%surf(t,y)%a_sin1(ireg%N))
	  allocate(wf%surf(t,y)%a_cos1(ireg%N))

      do i=1,ireg%N
        c=>ireg%coef(i)
        alfa=c%omega*time-c%k*(pos_h(1)*dcos(c%dir)+pos_h(2)*dsin(c%dir))+c%phi
        wf%surf(t,y)%a_sin1(i)=c%A*dsin(alfa)
		wf%surf(t,y)%eta=wf%surf(t,y)%eta+wf%surf(t,y)%a_sin1(i)
        wf%surf(t,y)%a_cos1(i)=c%A*dcos(alfa)
	  enddo
	enddo
	wd%time=0.d0
  enddo

  do i=1,ireg%N   ! loop over coef
    c=>ireg%coef(i)
    do z=1,wf%nz  ! loop over depth
      pos_h(3)=-wd%wdepth*(1.d0-dsin(pi*(z-1)/(2*(wf%nz-1))))
	       
      cosh1=dmin1(dcosh(c%k*(pos_h(3)+wd%wdepth)),1.0D308)
	  sinh1=dmin1(dsinh(c%k*(pos_h(3)+wd%wdepth)),1.0D308)
	  sinh2=dmin1(dsinh(c%k*wd%wdepth),1.0D308)
	  cosh2=dmin1(dcosh(c%k*wd%wdepth),1.0D308)

!	  write(0,*) 'k=',c%k
!	  write(0,*) 'pos_h(3)=',pos_h(3)
!	  write(0,*) cosh1,dcosh(c%k*(pos_h(3)+wd%wdepth)),cosh1-dcosh(c%k*(pos_h(3)+wd%wdepth))
!	  write(0,*) sinh1,dsinh(c%k*(pos_h(3)+wd%wdepth)),sinh1-dsinh(c%k*(pos_h(3)+wd%wdepth))
!	  write(0,*) sinh2,dsinh(c%k*wd%wdepth),sinh2-dsinh(c%k*wd%wdepth)
!	  write(0,*) cosh2,dcosh(c%k*wd%wdepth),cosh2-dcosh(c%k*wd%wdepth)

!      cosh1=dcosh(c%k*(pos_h(3)+wd%wdepth))
!	  sinh1=dsinh(c%k*(pos_h(3)+wd%wdepth))
!	  sinh2=dsinh(c%k*wd%wdepth)
!	  cosh2=dcosh(c%k*wd%wdepth)

      tmp1=c%omega*cosh1/sinh2
	  tmp2=c%omega*sinh1/sinh2
	  tmp3=c%omega**2*cosh1/sinh2
	  tmp4=c%omega**2*sinh1/sinh2
	  tmp5=wd%rho*wd%g*cosh1/cosh2

      do y=1,wf%ny ! loop over lateral positions

        do t=1,wf%nt ! loop over time
		  a_sin1=>wf%surf(t,y)%a_sin1(i)
		  a_cos1=>wf%surf(t,y)%a_cos1(i)
		  tmp=tmp1*a_sin1
          wf%kdata(t,z,y)%vel(1)=wf%kdata(t,z,y)%vel(1)+tmp*dcos(c%dir)
	      wf%kdata(t,z,y)%vel(2)=wf%kdata(t,z,y)%vel(2)+tmp*dsin(c%dir)
          wf%kdata(t,z,y)%vel(3)=wf%kdata(t,z,y)%vel(3)+tmp2*a_cos1

          tmp=tmp3*a_cos1
          wf%kdata(t,z,y)%acc(1)=wf%kdata(t,z,y)%acc(1)+tmp*dcos(c%dir)
	      wf%kdata(t,z,y)%acc(2)=wf%kdata(t,z,y)%acc(2)+tmp*dsin(c%dir)
          wf%kdata(t,z,y)%acc(3)=wf%kdata(t,z,y)%acc(3)-tmp4*a_sin1
          wf%kdata(t,z,y)%pres=wf%kdata(t,z,y)%pres+tmp5*a_sin1
		  ! these varables are no longer to be used
		  if ((i.eq.ireg%N).and.(z.eq.wf%nz).and.(y.eq.wf%ny)) then
            wf%surf(t,y)%a_sin1=>null()
            wf%surf(t,y)%a_cos1=>null()
		  endif

		enddo

	  enddo
	enddo
  enddo
  Write(0,*) 'In Wkin_dll, calculating wave field - done'
endif
end subroutine
!-------------------------------------------------------------

function Getvalue(xvec,yvec,N,q,switch)
use tools, only : intpol_r8
integer             :: N
real*8,dimension(N)	:: xvec
real*8,dimension(N) :: yvec
real*8				:: q
integer             :: switch !(0=result is yvec(q),1=result is xvec(q))
real*8				:: Getvalue

integer :: i

if (switch==1) then
  i=1
  do while (xvec(i)<q)
    i=1+1
	if (i==N) exit
  enddo
  getvalue=intpol_r8(q,xvec(i-1),xvec(i),yvec(i-1),yvec(i))

elseif (switch==2) then
  i=1
  do while (yvec(i)<q)
    i=i+1
	if (i==N) exit
  enddo
  getvalue=intpol_r8(q,yvec(i-1),yvec(i),xvec(i-1),xvec(i))
endif

end function Getvalue

!-------------------------------------------------------------

subroutine ireg_airy_sea_elevation(pos_h,ireg,elevation)
use tools
real*8,dimension(3)		:: pos_h
Type (Tireg)			:: ireg
real*8					:: elevation	

Type (Tcoef),pointer	:: c
integer					:: i

real*8					:: time			! look-up time
real*8					:: trel         ! relative time from time(t1)
integer					:: t1,t2,y1,y2
real*8					:: eps=1.d-7
real*8					:: yrel
Type (Twavefield),pointer	:: wf=>null()

if (ireg%pregen) then
  wf=>ireg%wfield
  time = wd%time - pos_h(1)/wf%groupvel

  t1=MAX(MIN(floor((time+wf%t0+eps)/wf%dt)+1,wf%nt),1)
  t2=MAX(MIN(floor((time+wf%t0+eps+wf%dt)/wf%dt)+1,wf%nt),1)
  trel=modulus_r_rea(time+wf%t0+eps,0.0d0,wf%dt)  

  if (wf%ny.eq.1) then
    elevation= intpol_r8(trel,0.d0,wf%dt,wf%surf(t1,1)%eta,wf%surf(t2,1)%eta)
  else
    y1=MIN(MAX(floor(pos_h(2)/wf%dy+(wf%ny-1)/2.d0+1.d0),1),wf%ny)
    y2=MIN(MAX(y1+1,1),wf%ny)
    elevation = intpol_area(trel,yrel,0.d0,0.d0,wf%dt,wf%dy, &
                          wf%surf(t1,y1)%eta,wf%surf(t2,y1)%eta, &
						  wf%surf(t2,y2)%eta,wf%surf(t1,y2)%eta)
  endif
else
  elevation=0
  do i=1,ireg%N
    c=>ireg%coef(i)
    elevation=elevation+c%A*dsin(c%omega*wd%time-c%k*(pos_h(1)*dcos(c%dir)+pos_h(2)*dsin(c%dir))+c%phi)   ! here elevation is positive upwards
  enddo
endif

end subroutine ireg_airy_sea_elevation

!-------------------------------------------------------------

subroutine ireg_airy_get_kinematics(ireg,pos_h,vel_h,acc_h,pres)
use tools
Type (Tireg)		:: ireg
real*8,dimension(3) :: pos_h
real*8,dimension(3) :: vel_h
real*8,dimension(3) :: acc_h
real*8				:: pres

real*8				:: tmp,tmp2
real*8				:: elevation,zwheeler
integer				:: i
Type (Tcoef),pointer	:: c
real*8				:: cosh1, cosh2, sinh1, sinh2, a_sin1, a_cos1
real*8					:: time			! look-up time
integer					:: t1,t2,y1,y2,z1,z2
real*8					:: trel,yrel,zrel ! relative time from time(t1)
real*8					:: eps=1.d-7
Type (Twavefield),pointer	:: wf=>null()
real*4,dimension(3)		:: F1,F2,F3,F4
!real*8,dimension(3)		:: F1,F2,F3,F4
real*8					:: z1pos,z2pos,dz
real*8					:: zlimit=0.d0    ! upper limit for wave kinematics calculation

vel_h=0.d0
acc_h=0.d0
pres=0.d0		! Dynamic pressure is calculated and summarized for all wave components

! Get elevation
call ireg_airy_sea_elevation(pos_h,ireg,elevation)		! Might be possible to increase speed

! lookup kinematics
select case (ireg%stretching)
case (0)
  zlimit=0.d0
case (1)
  zlimit=elevation
case default
end select

if ((pos_h(3).LE.zlimit).and.(pos_h(3).GE.-wd%wdepth)) then
  if (ireg%pregen) then
  !--------------------------------------------------
  ! Lookup in precalculated field
    wf=>ireg%wfield

    time = wd%time - pos_h(1)/wf%groupvel

    t1=MAX(MIN(floor((time+wf%t0+eps)/wf%dt)+1,wf%nt),1)
    t2=MAX(MIN(floor((time+wf%t0+eps+wf%dt)/wf%dt)+1,wf%nt),1)
    trel=modulus_r_rea(time+wf%t0+eps,0.0d0,wf%dt)  

    if (wf%ny.gt.1) then
      y1=MAX(MIN( floor(((wf%ny-1)/2.d0*wf%dy+floor(pos_h(2)/wf%dy)*wf%dy)/wf%dy) ,wf%nt),1)
      y2=MIN(MAX(y1+1,2),wf%ny)
	  yrel=((wf%ny-1)/2.d0*wf%dy+pos_h(2)-y1*wf%dy)/wf%dy
	else
	  y1=1
	  y2=1
	endif

    select case (ireg%stretching)
    case (0)  ! no stretching
      tmp2=DASIN(pos_h(3)/wd%wdepth+1.d0)*2.d0*(wf%Nz-1)/pi+1.d0
      z1=FLOOR(tmp2)
	  z1=MIN(z1,wf%nz)
	  z2=MIN(z1+1,wf%nz)
	  z1pos=-wd%wdepth*(1.d0-DSIN(pi*(z1-1)/(2.d0*(wf%Nz-1))))
	  z2pos=-wd%wdepth*(1.d0-DSIN(pi*(z2-1)/(2.d0*(wf%Nz-1))))
	  zrel=pos_h(3)-z1pos										! z distance between pos_h(3) and z1
	  dz=max(z2pos-z1pos,eps)
    case (1)	! wheeler stretching
      zwheeler=(pos_h(3)+wd%wdepth)/(1.0d0+elevation/wd%wdepth)
      tmp2=DASIN((zwheeler-wd%wdepth)/wd%wdepth+1.d0)*2.d0*(wf%Nz-1)/pi+1.d0
      z1=FLOOR(tmp2)
	  z2=MIN(z1+1,wf%nz)
	  z1pos=-wd%wdepth*(1.d0-DSIN(pi*(z1-1)/(2.d0*(wf%Nz-1))))
	  z2pos=-wd%wdepth*(1.d0-DSIN(pi*(z2-1)/(2.d0*(wf%Nz-1))))
	  zrel=zwheeler-wd%wdepth-z1pos
	  dz=max(z2pos-z1pos,eps)
    case default
    end select

    if (wf%ny.eq.1) then
      F1=wf%kdata(t1,z1,1)%vel
      F2=wf%kdata(t1,z2,1)%vel
      F3=wf%kdata(t2,z1,1)%vel
      F4=wf%kdata(t2,z2,1)%vel
	else
	  ! perform lookup here - velocity
      ! interpolate in the y direction - eliminate this variable
	  F1=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t1,z1,y1)%vel,wf%kdata(t1,z1,y2)%vel)
	  F2=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t1,z2,y1)%vel,wf%kdata(t1,z2,y2)%vel)
	  F3=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t2,z1,y1)%vel,wf%kdata(t2,z1,y2)%vel)
	  F4=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t2,z2,y1)%vel,wf%kdata(t2,z2,y2)%vel)
	endif
	! interpolate in the z direction - eliminate this variable
	F1=intpol_r84_v(zrel,0.d0,dz,F1,F2)
	F2=intpol_r84_v(zrel,0.d0,dz,F3,F4)
	! interpolate in the t direction - eliminate this variable
	vel_h=intpol_r84_v(trel,0.d0,wf%dt,F1,F2)*1.d0

    !-------------------------------------
    if (wf%ny.eq.1) then
      F1=wf%kdata(t1,z1,1)%acc
      F2=wf%kdata(t1,z2,1)%acc
      F3=wf%kdata(t2,z1,1)%acc
      F4=wf%kdata(t2,z2,1)%acc
	else
	  ! perform lookup here - acceleration
      ! interpolate in the y direction - eliminate this variable
	  F1=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t1,z1,y1)%acc,wf%kdata(t1,z1,y2)%acc)
	  F2=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t1,z2,y1)%acc,wf%kdata(t1,z2,y2)%acc)
	  F3=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t2,z1,y1)%acc,wf%kdata(t2,z1,y2)%acc)
	  F4=intpol_r84_v(yrel,0.d0,wf%dy,wf%kdata(t2,z2,y1)%acc,wf%kdata(t2,z2,y2)%acc)
	endif
	! interpolate in the z direction - eliminate this variable
	F1=intpol_r84_v(zrel,0.d0,dz,F1,F2)
	F2=intpol_r84_v(zrel,0.d0,dz,F3,F4)
	! interpolate in the t direction - eliminate this variable
	acc_h=intpol_r84_v(trel,0.d0,wf%dt,F1,F2)*1.d0

    !-------------------------------------
    if (wf%ny.eq.1) then
      F1(1)=wf%kdata(t1,z1,1)%pres
      F2(1)=wf%kdata(t1,z2,1)%pres
      F3(1)=wf%kdata(t2,z1,1)%pres
      F4(1)=wf%kdata(t2,z2,1)%pres
	else
	  ! perform lookup here - dynamic pressure
      ! interpolate in the y direction - eliminate this variable
	  F1(1)=intpol_r84(yrel,0.d0,wf%dy,wf%kdata(t1,z1,y1)%pres,wf%kdata(t1,z1,y2)%pres)
	  F2(1)=intpol_r84(yrel,0.d0,wf%dy,wf%kdata(t1,z2,y1)%pres,wf%kdata(t1,z2,y2)%pres)
	  F3(1)=intpol_r84(yrel,0.d0,wf%dy,wf%kdata(t2,z1,y1)%pres,wf%kdata(t2,z1,y2)%pres)
	  F4(1)=intpol_r84(yrel,0.d0,wf%dy,wf%kdata(t2,z2,y1)%pres,wf%kdata(t2,z2,y2)%pres)
	endif
	! interpolate in the z direction - eliminate this variable
	F1(1)=intpol_r84(zrel,0.d0,dz,F1(1),F2(1))
	F2(1)=intpol_r84(zrel,0.d0,dz,F3(1),F4(1))
	! interpolate in the t direction - eliminate this variable
	pres=intpol_r84(trel,0.d0,wf%dt,F1(1),F2(1))*1.d0

  else
  !--------------------------------------------------
  ! Traditional approach with full update
    do i=1,ireg%N
      c=>ireg%coef(i)
      select case (ireg%stretching)
      case (0)  ! no stretching
	    a_sin1=c%A*dsin(c%omega*wd%time-c%k*(pos_h(1)*dcos(c%dir)+pos_h(2)*dsin(c%dir))+c%phi)
	    a_cos1=c%A*dcos(c%omega*wd%time-c%k*(pos_h(1)*dcos(c%dir)+pos_h(2)*dsin(c%dir))+c%phi)

        cosh1=dcosh(c%k*(pos_h(3)+wd%wdepth))
	    sinh1=dsinh(c%k*(pos_h(3)+wd%wdepth))
	    sinh2=dsinh(c%k*wd%wdepth)
	    cosh2=dcosh(c%k*wd%wdepth)

        tmp=c%omega*cosh1/sinh2*a_sin1
        vel_h(1)=vel_h(1)+tmp*dcos(c%dir)
	    vel_h(2)=vel_h(2)+tmp*dsin(c%dir)
        vel_h(3)=vel_h(3)+c%omega*sinh1/sinh2*a_cos1

        tmp=c%omega**2*cosh1/sinh2*a_cos1
        acc_h(1)=acc_h(1)+tmp*dcos(c%dir)
	    acc_h(2)=acc_h(2)+tmp*dsin(c%dir)
        acc_h(3)=acc_h(3)-c%omega**2*sinh1/sinh2*a_sin1
        pres=pres+wd%rho*wd%g*cosh1/cosh2*a_sin1
      case (1)	! wheeler stretching

	    a_sin1=c%A*dsin(c%omega*wd%time-c%k*(pos_h(1)*dcos(c%dir)+pos_h(2)*dsin(c%dir))+c%phi)
	    a_cos1=c%A*dcos(c%omega*wd%time-c%k*(pos_h(1)*dcos(c%dir)+pos_h(2)*dsin(c%dir))+c%phi)

        zwheeler=(pos_h(3)+wd%wdepth)/(1.0d0+elevation/wd%wdepth)

        cosh1=dcosh(c%k*zwheeler)
	    sinh1=dsinh(c%k*zwheeler)
	    sinh2=dsinh(c%k*wd%wdepth)
	    cosh2=dcosh(c%k*wd%wdepth)

        tmp=c%omega*cosh1/sinh2*a_sin1
        vel_h(1)=vel_h(1)+tmp*dcos(c%dir)
	    vel_h(2)=vel_h(2)+tmp*dsin(c%dir)
        vel_h(3)=vel_h(3)+c%omega*sinh1/sinh2*a_cos1

        tmp=c%omega**2*cosh1/sinh2*a_cos1
        acc_h(1)=acc_h(1)+tmp*dcos(c%dir)
	    acc_h(2)=acc_h(2)+tmp*dsin(c%dir)
        acc_h(3)=acc_h(3)-c%omega**2*sinh1/sinh2*a_sin1
        pres=pres+wd%rho*wd%g*cosh1/cosh2*a_sin1
      case default
      end select
    enddo
  endif
endif

end subroutine ireg_airy_get_kinematics
!-------------------------------------------------------------

function dispersion(k,omega,d)
use global_typesnvars, only : wd
real*8		:: k		! wave number
real*8		:: omega	! frequency [rad/s]
real*8		:: d		! water depth
real*8		:: dispersion ! function output - target is zero

dispersion=wd%g*k*dtanh(k*d)-omega**2

end function dispersion

!-------------------------------------------------------------
function dispersion_deriv(k,omega)
use global_typesnvars, only : wd
real*8		:: k		! wave number
real*8		:: omega	! frequency [rad/s]
real*8		:: dispersion_deriv	! function output

real*8,save	:: eps=1.0D-4

dispersion_deriv=(dispersion(k+eps,omega,wd%wdepth)-dispersion(k-eps,omega,wd%wdepth))/(2.0d0*eps)
end function dispersion_deriv

!-------------------------------------------------------------
subroutine jonswap(js_var,omega,Sp)
use global_typesnvars
Type (Tjonswap_var)	:: js_var
real*8			:: omega
real*8			:: Sp

real*8			:: omega_p
real*8			:: sigma
real*8			:: alfa

if (omega>0.0d0) then
!  Tp=js_var%Tz/dsqrt((5.0d0+js_var%gamma)/(11.0d0+js_var%gamma))
  omega_p=2.0d0*pi/js_var%Tp
  if (omega>omega_p) then
    sigma=0.07
  else
    sigma=0.09
  endif
  alfa=0.3125d0*js_var%Hs**2*omega_p**4/wd%g**2*(1.0d0-0.287*dlog(js_var%gamma))
  Sp=alfa*wd%g**2/omega**5*dexp(-1.25d0*(omega_p/omega)**4)* &
     js_var%gamma**(dexp((omega-omega_p)**2/(-2.0d0*sigma**2*omega_p**2)))
else
  Sp=0.d0
endif
end subroutine jonswap

!-------------------------------------------------------------
subroutine pm(pm_var,omega,Sp)
use global_typesnvars
Type (Tpm_var)	:: pm_var
real*8			:: omega
real*8			:: Sp

if (omega>0.0d0) then
  Sp=pm_var%Hs**2*pm_var%Tz/(8.d0*pi**2)*(omega*pm_var%Tz/(2.d0*pi))**(-5)* &
     dexp(-1.d0/pi*(omega*pm_var%Tz/(2*pi))**(-4))
else
  Sp=0.d0
endif
end subroutine pm
end module ireg_airy