module ireg_airy_types

implicit none

!-----------------------------------------------
Type Tcoef
  real*8								::	A		=0.d0		! Amplitude
  real*8								::	omega	=0.d0	! omega (frequency) [rad/s]
  real*8								::	phi		=0.d0		! phase
  real*8								::	k		=0.d0		! wave number
  real*8								::  dir     =0.d0       ! wave direction
End type Tcoef

!-----------------------------------------------
Type Tjonswap_var
  real*8	:: Hs		=0.d0			! Significant wave height
  real*8	:: Tp		=0.d0			! Peak frequency
  real*8	:: gamma	=0.d0			! Peak parameter
end type Tjonswap_var
!-----------------------------------------------
Type TPM_var
  real*8	:: Hs		=0.d0			! Significant wave height
  real*8	:: Tz		=0.d0			! Zero cross time period [s]
end type TPM_var
!-----------------------------------------------
Type TWN_var
  real*8	:: var_target=0.d0			! Target variance
  real*8	:: f0		=0.d0			! Minimum cutoff frequency [Hz]
  real*8	:: f1		=0.d0			! Maximum cutoff frequency [Hz]
end type TWN_var
!-----------------------------------------------
Type Tkindata
!  real*8,dimension(3)	:: vel=0.d0
!  real*8,dimension(3)	:: acc=0.d0
!  real*8	:: pres		=0.d0
  real*4,dimension(3)	:: vel=0.d0
  real*4,dimension(3)	:: acc=0.d0
  real*4	:: pres		=0.d0
end Type Tkindata
!-----------------------------------------------
Type Tsurface
  real*8	:: eta		=0.d0
  real*8,dimension(:),pointer	:: a_sin1=>null() ! tmp value
  real*8,dimension(:),pointer	:: a_cos1=>null() ! tmp value
!  real*4,dimension(:),pointer	:: a_sin1=>null() ! tmp value
!  real*4,dimension(:),pointer	:: a_cos1=>null() ! tmp value
end Type Tsurface
!-----------------------------------------------
Type Twavefield
  integer   :: nz		=15
  integer	:: nt		=0
  integer	:: ny		=1              ! field dimensions
  real*8	:: dt		=0.d0			! time interval
  real*8	:: dy		=0.d0			! lateral distance in grid
  real*8	:: xrange   =1.0d2			! range for xposition lookup (used to increase length of wavefield)
  real*8	:: t0		=0.d0			! Start time for lookup in wavefield (depends on xrange and groupvelocity only)

  real*8	:: groupvel = 0.d0			! wave field group velocity based on the largest wave component
  Type (Tkindata),dimension(:,:,:),pointer	:: kdata=>null() ! field data (format timestep,z_nr,y_nr)
  Type (Tsurface),dimension(:,:),pointer	:: surf=>null()  ! surface data (format timestep,y_nr)
end type Twavefield
!-----------------------------------------------
Type Tireg
  integer								::  stretching	=1 ! 0=nothing,1=wheeler
  integer								::  spectrum	=1  ! 1=jonswap
  integer								::  spreading   =0 ! Directional spreading (0=none, 1=on)
  integer                               ::  s           =4 ! Directional spreading s parameter
  Type (Tjonswap_var),pointer			::  jonswap		=>null()	! Jonswap parameters
  Type (TPM_var),pointer				::  PM			=>null()	! Pierson Moscowitz parameters
  Type (TWN_var),pointer				::  WN			=>null()	! White noise parameters
  real*8								::  omega_min	=0.d0 ! minimum frquency used [rad/s]
  real*8								::  omega_step	=0.d0 ! delta omega [rad/s]
  integer								::  N=1000		! number of coefficients
  Type (Tcoef),dimension(:),pointer		::	coef=>null()
  integer								::  seednr=1	! seed number (between 1 and 2147483646)

  logical								::  pregen=.true.	! True if wfield is pregenerated, false if old method is used
  Type (Twavefield),pointer				::  wfield=>null() ! pregenereated wave field
end type Tireg

end module ireg_airy_types