module stream_func
! Stream function wave kinematics
! Code from CW263, Chaplin
use global_typesnvars

implicit none

public :: stream_func_input
public :: stream_func_init
public :: stream_func_sea_elevation
public :: stream_func_get_kinematics

contains

!-------------------------------------------------------------

Subroutine stream_func_input(masterfile,strf)
use genout_tools
!use global_vars
use global_typesnvars, only : wd

Type (Tmasterfile), pointer :: masterfile
Type (Tstrf),pointer		:: strf

Type (Tline) :: line
Type (Tword) :: words

real*8, dimension(maxwords) :: rea

logical :: init, &
           out, &
           err, &
           all_ok, &
		   stretching_control, &
		   wave_control

init=.false.
out=.false.
stretching_control=.false.
wave_control=.false.

do while (out==.false.)
  call readline(masterfile,line,init)        ! line is read
    call getwords(line%line,words,masterfile)                           ! words in line is retrieved
    select case (trim(words%word_vec(1)))
    
	case ('wave')
      call get_parameters(words,2,3,rea,err)
      if (err==.true.) then; goto 10; endif
      strf%Hs=rea(1)
	  if (rea(2)>0.d0) then
        strf%T=rea(2)
	  else
	    write(0,*) '*** ERROR *** In strf. airy input, T less or equal to 0.d0'
	  endif
	  strf%current=rea(3)

      wave_control=.true.

    case ('')
      continue

    case ('end')
      out=.true.
      ! Control of data given above is OK
      all_ok=.true.
      if ((wave_control==.false.)) then
        all_ok=.false.
      endif
      if (all_ok==.true.) then
        write(0,*) 'Reg. airy input commands read with succes'
      else
        write(0,*) '*** ERROR *** Not all needed wkin_dll stream_func input commands present - error'
      endif
    case (';')
      continue
    case default
      write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    end select
enddo
return
10  write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)

end subroutine stream_func_input

!-------------------------------------------------------------

subroutine stream_func_init(strf)
Type (Tstrf)	:: strf

real*4 :: d,T,Hs,c,el
integer n
d=wd%wdepth
T=strf%T
Hs=strf%Hs
c=strf%current
call cw260(d,T,Hs,c,1,n,el)
strf%n=n
strf%el=el
!call cw260(wd%wdepth*1.0,strf%T*1.0,strf%Hs*1.0,strf%current*1.0,1,strf%n,strf%el)

end subroutine stream_func_init

!-------------------------------------------------------------

subroutine stream_func_sea_elevation(pos_h,strf,elevation)
real*8,dimension(3)		:: pos_h
Type (Tstrf)				:: strf
real*8					:: elevation	

real*4 :: xx,yy,tt,uu,vv,ut,vt,du,dv,etah
integer :: n

n=strf%n
xx=pos_h(1)
!yy=pos_h(2)
yy=pos_h(3)
tt=wd%time
call kmts(n,xx,yy,tt,uu,vv,ut,vt,du,dv,etah)
elevation=etah

end subroutine stream_func_sea_elevation

!-------------------------------------------------------------

subroutine stream_func_get_kinematics(strf,pos_h,vel_h,acc_h,pres)
Type (Tstrf)			:: strf
real*8,dimension(3) :: pos_h
real*8,dimension(3) :: vel_h
real*8,dimension(3) :: acc_h
real*8				:: pres

real*4 :: xx,yy,tt,uu,vv,ut,vt,du,dv,etah
integer :: n

n=strf%n
xx=pos_h(1)
!yy=pos_h(2)
yy=pos_h(3)
tt=wd%time
call kmts(n,xx,yy,tt,uu,vv,ut,vt,du,dv,etah)

vel_h(1)=uu
vel_h(2)=0.d0
vel_h(3)=vv

acc_h(1)=du
acc_h(2)=0.d0
acc_h(3)=dv

pres=0.d0  ! limitation so far

end subroutine stream_func_get_kinematics

!-------------------------------------------------------------

end module stream_func