module tools
implicit none

TYPE Tfirstordervar
  real*8 :: tau=0.0d0,&
			 y1=0.0d0, &
			 y1_old=0.0d0
  logical :: init=.true.
END TYPE Tfirstordervar 

contains

function modulus_r_rea(x,x0,x1)
! modulus in range
real*8 :: x   ! input
real*8 :: x0  ! lower limit
real*8 :: x1  ! upper limit
real*8 :: modulus_r_rea

modulus_r_rea=modulo(x,x1-x0)+x0

end function modulus_r_rea
!---------------------------------------------------
function modulus_r_int(x,x0,x1)
! modulus in range
integer :: x   ! input
integer :: x0  ! lower limit (x>=x0)
integer :: x1  ! upper limit (x<x1)
integer :: modulus_r_int

modulus_r_int=modulo(x-x0,x1-x0+1)+x0

end function modulus_r_int

!---------------------------------------------------

Function intpol_r8(x,x0,x1,f0,f1)   ! Linear interpolation function                                     
  real*8 intpol_r8
  real*8,intent(in):: x,x0,x1,f0,f1
  intpol_r8=(x-x1)/(x0-x1)*f0+(x-x0)/(x1-x0)*f1
end function intpol_r8

!---------------------------------------------------

Function intpol_r84(x,x0,x1,f0,f1)   ! Linear interpolation function                                     
  real*8 intpol_r84
  real*8,intent(in):: x,x0,x1
  real*4,intent(in):: f0,f1
  intpol_r84=(x-x1)/(x0-x1)*f0+(x-x0)/(x1-x0)*f1
end function intpol_r84

!---------------------------------------------------

Function intpol_r8_v(x,x0,x1,f0,f1)   ! Linear interpolation between one dim arrays
  real*8,intent(in)             :: x,x0,x1   
  real*8,dimension(:),intent(in):: f0,f1
  real*8,dimension(1:size(f0))  :: intpol_r8_v

  integer i
  do i=1,size(f0)
    intpol_r8_v(i)=intpol_r8(x,x0,x1,f0(i),f1(i))
  enddo
end function intpol_r8_v

!---------------------------------------------------

Function intpol_r84_v(x,x0,x1,f0,f1)   ! Linear interpolation between one dim arrays
  real*8,intent(in)             :: x,x0,x1   
  real*4,dimension(:),intent(in):: f0,f1
  real*4,dimension(1:size(f0))  :: intpol_r84_v

  integer i
  do i=1,size(f0)
    intpol_r84_v(i)=intpol_r84(x,x0,x1,f0(i),f1(i))
  enddo
end function intpol_r84_v

!---------------------------------------------------

Function intpol_area(x,y,x0,y0,dx,dy,f0,f1,f2,f3)   ! Linear interpolation function based on area calculations                                    
  real*8 intpol_area
  real*8,intent(in):: x,y,x0,y0,dx,dy,f0,f1,f2,f3
  real*8 :: V0,V1,V2,V3,Vtot
! F0 are at x0,y0
! F1 are at x0+dx,y0
! F2 are at x0+dx,y0+dy
! F3 are at x0,y0+dy

  V0=(x-x0)*(y-y0)
  V1=(x0+dx-x)*(y-y0)
  V2=(x0+dx-x)*(y0+dy-y)
  V3=(x-x0)*(y0+dy-y)
  Vtot=dx*dy

  intpol_area=(V0*f2+V1*f3+V2*f0+V3*f1)/Vtot
end function intpol_area

!---------------------------------------------------

Function intpol_area_vec(x,y,x0,y0,dx,dy,f0,f1,f2,f3)   ! Linear interpolation function based on area calculations, where f now are vectors                                    
  real*8,intent(in):: x,y,x0,y0,dx,dy
  real*8,dimension(:),intent(in) :: f0,f1,f2,f3
  real*8 :: V0,V1,V2,V3,Vtot
  real*8,dimension(size(f0)) :: intpol_area_vec
! F0 are at x0,y0
! F1 are at x0+dx,y0
! F2 are at x0+dx,y0+dy
! F3 are at x0,y0+dy

  V0=(x-x0)*(y-y0)
  V1=(x0+dx-x)*(y-y0)
  V2=(x0+dx-x)*(y0+dy-y)
  V3=(x-x0)*(y0+dy-y)
  Vtot=dx*dy

  intpol_area_vec=(V0*f2+V1*f3+V2*f0+V3*f1)/Vtot
end function intpol_area_vec

!---------------------------------------------------

Function intpol_vol_cart(x,y,z,x0,y0,z0,dx,dy,dz,F0,F1,F2,F3,F4,F5,F6,F7)
! Linear interpolation in cartesian box by use of volume weight functions
! x0,y0,z0 are position of lowest indexed corner of box
! dx,dy,dz are side lenghts of the box
! F0,F1,F2,..,F7 are point values at the corners
! F0 are at x0,y0,z0
! F1 are at x0,y0,z0+dz
! F2 are at x0,y0+dy,z0+dz
! F3 are at x0,y0+dy,z0
! F4 are at x0+dx,y0,z0
! F5 are at x0+dx,y0,z0+dz
! F6 are at x0+dx,y0+dy,z0+dz
! F7 are at x0+dx,y0+dy,z0
real*4,intent(in) :: x,y,z,x0,y0,z0,dx,dy,dz,F0,F1,F2,F3,F4,F5,F6,F7
real*4               :: intpol_vol_cart

real*4 :: V0,V1,V2,V3,V4,V5,V6,V7,Vtot
! calculation of volumes
V0=abs((x-x0)*(y-y0)*(z-z0))
V1=abs((x-x0)*(y-y0)*(z0+dz-z))
V2=abs((x-x0)*(y0+dy-y)*(z0+dz-z))
V3=abs((x-x0)*(y0+dy-y)*(z-z0))

V4=abs((x0+dx-x)*(y-y0)*(z-z0))
V5=abs((x0+dx-x)*(y-y0)*(z0+dz-z))
V6=abs((x0+dx-x)*(y0+dy-y)*(z0+dz-z))
V7=abs((x0+dx-x)*(y0+dy-y)*(z-z0))

Vtot=abs(dx*dy*dz)

intpol_vol_cart=(F0*V6+F1*V7+F2*V4+F3*V5+ &
                 F4*V2+F5*V3+F6*V0+F7*V1)/Vtot
end function intpol_vol_cart


!--------------------------------------------------------------------
!--------------------------------------------------------------------

Function determinant(A1,A2,B1,B2)
real*8, intent(in) :: A1,A2,B1,B2
real*8             :: determinant

determinant=A1*B2-A2*B1
end function determinant

!--------------------------------------------------------------------
!--------------------------------------------------------------------

Function cross(A,B)
real*8, dimension(3), intent(in) :: A,B
real*8, dimension(3) :: cross

cross(1)=determinant(A(2),A(3),B(2),B(3))
cross(2)=determinant(A(3),A(1),B(3),B(1))
cross(3)=determinant(A(1),A(2),B(1),B(2))
end function cross

!---------------------------------------------------

Function rotate(a,n,azi)
! Function that rotates the vector a around another vector n
! with the angle azi
! a and n both intersect the same base point of a
real*8,dimension(1:3),intent(in) :: a,n
real*8,intent(in)                :: azi
real*8,dimension(1:3)            :: rotate

rotate=n*(dot_product(n,a))+ &
       dcos(azi)*(a-n*(dot_product(n,a)))+ &
	   dsin(azi)*(cross(n,a))
end function rotate

!---------------------------------------------------

Function newton_raphson(x0,y,z,F,df,limit)
! Function that solves an equation using newton raphson approach

  interface
    function F(x,y,z)    ! user defined function
	  real*8 :: x,y,z
	  real*8 :: F
    end function F
  end interface

  interface
    function df(x,y)   ! the user differentiated function of F
	  real*8 :: x,y
	  real*8 :: df
    end function df
  end interface

  real*8,intent(in)  :: x0, &      ! starting point (best guess)
                        y,z
  real*8, intent(in) :: limit   ! convergence criteria
  real*8             :: newton_raphson

  real*8 :: residual, &
            x_n, &
			Fu, &
			dFu

  integer :: i

  residual=2.0d0*limit
  x_n=x0
  i=0
  do while (residual>limit)
    i=i+1
	Fu=f(x_n,y,z)
	dFu=df(x_n,y)
    newton_raphson=x_n-Fu/dFu
    residual=abs(newton_raphson-x_n)
	x_n=newton_raphson
	if (i>100) then
	  write(0,*) 'More that 100 iterations in newton raphson iteration of induction factor'
	  exit
	endif
  enddo
end function newton_raphson

!---------------------------------------------------
! Integration of an array based on simpsons rule
Function simpson_int(a,b,m,f)
real*8		:: a
real*8		:: b
integer		:: m
real*8,dimension(0:2*m)		:: f
real*8		:: simpson_int

real*8	:: s0,s1,s2,h
integer	:: i

  s0=f(1)+f(2*m)
  s1=0.d0
  do i=1,m
    s1=s1+f(i*2-1)
  enddo
  s2=0.d0
  do i=1,m-1
    s2=s1+f(i*2)
  enddo
  h=(b-a)/(2*m)
  simpson_int=h/3.0d0*(s0+4.d0*s1+2.d0*s2)
end function simpson_int

Function Firstorderfilt(firstc,deltat,Fvar,x)
! first order filter function based on indicial function expression
  Type (Tfirstordervar),intent(inout)	:: Fvar				! Filter variables
  real*8, intent(in)					:: deltat			! time step 
  real*8, intent(in)					:: x				! ingoing variable
  logical,intent(in)					:: firstc			! True in first call of time step
  real*8								:: Firstorderfilt	! result (filtered variable)

  real*8 :: exp_val
  real*8 :: y

  if (Fvar%init) then
	Fvar%y1=x
!	Fvar%y1=0.d0
	Fvar%init=.false.
  endif

  if (firstc) then
    Fvar%y1_old=Fvar%y1
  endif

  exp_val=-deltat/Fvar%tau
  if (exp_val<-100.0d0) then
    exp_val=0.0d0
  else
    exp_val=exp(exp_val)
  endif

  y=Fvar%y1_old*exp_val+x*(1.0d0-exp_val)

  Fvar%y1=y
  Firstorderfilt=y

end function Firstorderfilt

end module tools