module wavemods
use global_typesnvars
use tools
use wavemods_types
use NRiF

implicit none

public :: wavemods_input
public :: wavemods_init
public :: wavemods_sea_elevation
public :: wavemods_get_kinematics

contains

!-------------------------------------------------------------

Subroutine wavemods_input(masterfile,wavemods)
use genout_tools
!use global_vars
use global_typesnvars, only : wd

Type (Tmasterfile), pointer :: masterfile
Type (Twavemods),   pointer :: wavemods

Type (Tline) :: line
Type (Tword) :: words

real*8, dimension(maxwords) :: rea

logical :: init, &
           out, &
           err, &
           all_ok, &
           datafile1, & 
		   datafile2, & 
		   N1, & 
		   N2, & 
           Nd , &
           Nt

init=.false.
out=.false.
datafile1=.false.
datafile2=.false.
Nd=.false.
Nt=.false.
N1=.false.
N2=.false.

! note that wavetype is not known - data should just be read into wd%wavetype structure

do while (out==.false.)
  call readline(masterfile,line,init)         ! line is read
    call getwords(line%line,words,masterfile) ! words in line is retrieved
    select case (trim(words%word_vec(1)))

	case ('datafile_y')
      if (words%nwords>1) then 
       datafile1 = .true.
       wavemods%datafile1 = trim(words%word_vec(2))
      end if

	case ('datafile_z')
      if (words%nwords>1) then 
       datafile2 = .true.
       wavemods%datafile2 = trim(words%word_vec(2))
      end if

	case ('datafile_nd')
      call get_parameters(words,2,1,rea,err)
      if (err==.false.) then
       Nd=.true.
       wavemods%Nd=nint(rea(1))
      endif
      
	case ('datafile_nt')
      call get_parameters(words,2,1,rea,err)
      if (err==.false.) then
       Nt=.true.
       wavemods%Nt=nint(rea(1))
      endif

	case ('datafile_t0')
      call get_parameters(words,2,1,rea,err)
      if (err==.false.) then
       wavemods%filet0=rea(1)
      endif

	case ('datafile_depth')
      call get_parameters(words,2,1,rea,err)
      if (err==.false.) then
       wavemods%zmin=-rea(1)
      endif

    case ('ncol_y')
      call get_parameters(words,2,1,rea,err)
      if (err==.false.) then
	   N1=.true.
       wavemods%Ndum1=rea(1)
      endif

	case ('ncol_z')
      call get_parameters(words,2,1,rea,err)
      if (err==.false.) then
	   N2=.true.
       wavemods%Ndum2=rea(1)
      endif

    case ('')
      continue

    case ('end')
      out=.true.
      ! Control of data given above is OK
      all_ok=.true.
      if ((datafile1==.false.) .or. (datafile2==.false.) .or. (Nt==.false.) .or. (Nd==.false.))  then
       all_ok=.false.
      endif
      if (all_ok==.true.) then
        write(0,*) 'wavemods input commands read with succes'
      else
        write(0,*) '*** ERROR *** Not all needed wkin_dll wavemods input commands present - error'
      endif
    case (';')
      continue
    case default
      write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)
    end select
enddo

! screen output
write (*, '(" wkin_dll:wavemods_input:datafile1",a,":")') trim(wavemods%datafile1)
write (*, '(" wkin_dll:wavemods_input:datafile2",a,":")') trim(wavemods%datafile2)

return
10  write(0,*) '*** ERROR *** Error in command line ',masterfile%linenr,' Masterfile name: ',trim(masterfile%name)



end subroutine wavemods_input

!-------------------------------------------------------------

subroutine wavemods_init(wavemods,t0,t1,dt)
use tools
use msimsl
Type (Twavemods)	:: wavemods
real*8		 	    :: t0
real*8			    :: t1
real*8			    :: dt
character*2048      :: line = ' '
logical :: fnexist
integer :: i, j
real*8,dimension(:),allocatable :: dum_u
real*8,dimension(:),allocatable :: dum_w
real*8,dimension(:),allocatable :: dep
integer :: Ndep, Ndum1, Ndum2

Ndum1 = wavemods%Ndum1
Ndum2 = wavemods%Ndum2
Ndep = wavemods%Nd

allocate(dum_u(Ndum1))
allocate(dum_w(Ndum2))
allocate(dep(Ndep))

! allocate space ************************ where to free ? **************************************
allocate(wavemods%z(wavemods%nt,wavemods%nd))
allocate(wavemods%u(wavemods%nt,wavemods%nd))
allocate(wavemods%w(wavemods%nt,wavemods%nd))
allocate(wavemods%d2u_dz2(wavemods%nt,wavemods%nd))
allocate(wavemods%a(wavemods%nt,wavemods%nd))
allocate(wavemods%aw(wavemods%nt,wavemods%nd))
allocate(wavemods%d2a_dz2(wavemods%nt,wavemods%nd))
allocate(wavemods%t(wavemods%nt))
allocate(wavemods%eta(wavemods%nt))
allocate(wavemods%d2eta_dt2(wavemods%nt))


wavemods%z = 0.000e0

  ! check if datafile exists
  inquire(file = wavemods%datafile1, exist = fnexist)
  if (fnexist==.false.) then
   write (*, '("*** ERROR *** file note found:",a)') wavemods%datafile1
   stop
  endif
  
  inquire(file = wavemods%datafile2, exist = fnexist)
  if (fnexist==.false.) then
   write (*, '("*** ERROR *** file note found:",a)') wavemods%datafile2
   stop
  endif
  

  ! read data file into matrices
  write (*,*) ' wkin_dll:wavemods_init: reading datafile'

  ! This is reading U components
  open(10, file=wavemods%datafile1, status='old', readonly)
  
     ! rading 9 headerlines 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, '(a)') line 
   read (10, *) (dep(i),i=1,wavemods%Nd) 
   read (10, '(a)') line
   do i=1,wavemods%Nt
     ! read row
     read (10,*) (dum_u(j),j=1,Ndum1)
     ! distribute data from row
     wavemods%t(i)    = dum_u(1)
     wavemods%eta(i)  = dum_u(2)
     do j=1, wavemods%Nd
       wavemods%u(i,j) = dum_u(2+j)
     enddo
     do j=1, wavemods%Nd
       wavemods%a(i,j) = dum_u(2+wavemods%Nd+j)
     enddo
	 do j=1, wavemods%Nd
       wavemods%z(i,j) = dep(j)*(wavemods%eta(i)-wavemods%zmin)+wavemods%zmin
     enddo	 
   enddo
  close(10)

! This is reading W components
   open(100, file=wavemods%datafile2, status='old', readonly)
   ! rading 9 headerlines 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, '(a)') line 
   read (100, *) (dep(i),i=1,wavemods%Nd) 
   read (100, '(a)') line
   do i=1,wavemods%Nt
     ! read row
     read (100,*) (dum_w(j),j=1,Ndum2)
     do j=1, wavemods%Nd
       wavemods%w(i,j) = dum_w(1+j)
     enddo
     do j=1, wavemods%Nd
       wavemods%aw(i,j) = dum_w(1+wavemods%Nd+j)
     enddo	 
   enddo
  close(100)

   write (*, '("  wkin_dll:wavemods_init: data read for ",i," discrete times")') i-1
   wavemods%tmax = wavemods%t(i-1)
!   wavemods%zmin = wavemods%z(1,1) ! <----------------------------------------------- Note zmin must be time independent in data file (checked below)

  ! check validity of input
  do i=1,wavemods%Nt
   if (wavemods%z(i,1) /= wavemods%zmin) then 
    write (*,*) '*** ERROR *** min depth in datafile is not constant for all discrete times. t=' 
    write (*,*) wavemods%t(i) 
   end if
  enddo
  
deallocate(dum_u)
deallocate(dum_w)
deallocate(dep)

end subroutine wavemods_init

!-------------------------------------------------------------

subroutine wavemods_sea_elevation(pos_h,wavemods,elevation)
real*8,dimension(3)		:: pos_h
Type (Twavemods)		:: wavemods
real*8					:: elevation	
real*8                  :: t
integer                 :: n0, n1

  t  = wd%time+wavemods%filet0
  n0 = floor(t/wavemods%tmax*(wavemods%Nt-1) + 1)
  n1 = n0 + 1
  elevation = intpol_r8(t,wavemods%t(n0),wavemods%t(n1),wavemods%eta(n0),wavemods%eta(n1))

end subroutine wavemods_sea_elevation

!-------------------------------------------------------------

subroutine wavemods_get_kinematics(wavemods,pos_h,vel_h,acc_h,pres)
! waves are assumed to travel in x-direction. 
! Vertical motion not included
! pres=hydrostatic pressure from water - no atmospheric pressure added
! 
Type (Twavemods)    :: wavemods
real*8,dimension(3) :: pos_h
real*8,dimension(3) :: vel_h
real*8,dimension(3) :: acc_h
real*8				:: pres
real*8              :: z  ! actual interpolation position to use, function of pos_h
real*8              :: u  ! x velocity
real*8              :: a  ! x accelation
real*8              :: w  ! z velocity
real*8              :: aw ! z accelation
real*8              :: t, t0, t1
!real*8              :: Dz, Dz0, Dz1, y, z0, z1, u0, u1, a0, a1
real*8              :: Dz, y, z0, z1, u0, u1, a0, a1
real*8              :: w0, w1, aw0, aw1
real*8              :: elevation
integer             :: nx0,nx1,nz00,nz01,nz10,nz11
u    = 0.0d0
a    = 0.0d0
w    = 0.0d0
aw   = 0.0d0
pres = 0.0d0

 ! this is the time used for lookup in data
 t = wd%time + wavemods%filet0

 ! this may cause overhead but elevation is calculated here to make sure it corresponds to correct time
 call wavemods_sea_elevation(pos_h,wavemods,elevation)

 z = pos_h(3) ! z is positive up

 ! check if evaluation position is between zmin and water surface
 if ((wavemods%zmin < z) .and. (z < elevation)) then

   ! linear interpolation - allows varying z-spacing for varying elevation

   ! relevant time indices
   nx0 = floor(t/wavemods%tmax*(wavemods%Nt-1) + 1)
   nx1 = nx0 + 1
   t0  = wavemods%t(nx0)
   t1  = wavemods%t(nx1)

   ! data
   Dz  = elevation - wavemods%zmin
   y   = (z - wavemods%zmin)/Dz
   z0  = wavemods%zmin + y*(wavemods%eta(nx0) - wavemods%zmin)
   z1  = wavemods%zmin + y*(wavemods%eta(nx1) - wavemods%zmin)

   ! relevant z indices
   call LOCATE(wavemods%z(nx0,:),wavemods%Nd,z0,nz00)
   call LOCATE(wavemods%z(nx1,:),wavemods%Nd,z1,nz10)
   nz01 = nz00 + 1
   nz11 = nz10 + 1
   if (locate_checkbounds(wavemods%Nd,nz00)) then 
    write (*,*) ' *** ERROR *** locate is out of array bounds, paused'
    pause
   end if
   if (locate_checkbounds(wavemods%Nd,nz10)) then 
    write (*,*) ' *** ERROR *** locate is out of array bounds, paused'
    pause
   end if

   ! interpolation in z direction 
   u0   = intpol_r8(z0,wavemods%z(nx0,nz00),wavemods%z(nx0,nz01),wavemods%u(nx0,nz00),wavemods%u(nx0,nz01))
   a0   = intpol_r8(z0,wavemods%z(nx0,nz00),wavemods%z(nx0,nz01),wavemods%a(nx0,nz00),wavemods%a(nx0,nz01))
   u1   = intpol_r8(z1,wavemods%z(nx1,nz10),wavemods%z(nx1,nz11),wavemods%u(nx1,nz10),wavemods%u(nx1,nz11))
   a1   = intpol_r8(z1,wavemods%z(nx1,nz10),wavemods%z(nx1,nz11),wavemods%a(nx1,nz10),wavemods%a(nx1,nz11))

   w0   = intpol_r8(z0,wavemods%z(nx0,nz00),wavemods%z(nx0,nz01),wavemods%w(nx0,nz00),wavemods%w(nx0,nz01))
   aw0  = intpol_r8(z0,wavemods%z(nx0,nz00),wavemods%z(nx0,nz01),wavemods%aw(nx0,nz00),wavemods%aw(nx0,nz01))
   w1   = intpol_r8(z1,wavemods%z(nx1,nz10),wavemods%z(nx1,nz11),wavemods%w(nx1,nz10),wavemods%w(nx1,nz11))
   aw1  = intpol_r8(z1,wavemods%z(nx1,nz10),wavemods%z(nx1,nz11),wavemods%aw(nx1,nz10),wavemods%aw(nx1,nz11))

   ! interpolation in time
   u = intpol_r8(t,t0,t1,u0,u1)
   a = intpol_r8(t,t0,t1,a0,a1)

   w  = intpol_r8(t,t0,t1,w0,w1)
   aw = intpol_r8(t,t0,t1,aw0,aw1)

   pres    = 0.0d0

 end if

vel_h(1)=u
vel_h(2)=0.0d0
vel_h(3)=w  ! z is up
 
acc_h(1)=a
acc_h(2)=0.0d0
acc_h(3)=aw  ! z is up

end subroutine wavemods_get_kinematics

!-------------------------------------------------------------

end module wavemods