

subroutine init(inputfile,t0,t1,dt)
use global_typesnvars
use dfport
use ireg_airy
use reg_airy
use det_airy
use stream_func
!use wavemods
use master_input , only : read_inputfile

implicit none
character*(*) :: inputfile
real*8			:: t0	! start time for simulation
real*8			:: t1	! stop time for simulation
real*8			:: dt	! time increment

character*5		:: version
!DEC$ ATTRIBUTES DLLEXPORT, ALIAS:'init'::init


! Version info
! 1-0	TJUL	Basic edition by TJUL
! 1-1	ANMH	Wave field can be read by file and used directly through fft conversion
! 1-2	TJUL	Directional spreading included
! 1-3   ANMH	Bug corrected regarding read on seed number using iregular waves
! 1-4   TJUL	Pierson-Moscowitz spectrum added as option
!				Stream function wave added
!				Possible pre processing of wave field to speed up simulation time and enable many more coeffients				
! 1-5   TJUL	Bug in stream function wave. Static pressure was included - now removed
! 1-6	TJUL	Bug in stream fuction wave. lateral position was applied instead of vertical in kinematics look-up!!!
! 1-7	TKIM	New wave format for precalculated (high order) wave fields
! 1-8   ANMH	Update in deterministic iregular waves+bugfix
!version='1-9'	!   TJUL	New option for white noise wave exitation
version='2-0'	!   TJUL	Bug fix of version 1-9. Version 1-9 had some debug statements included that could meas up the time.


print*,'In wkin_dll init, inputfile: ',inputfile
print*,'Wkin.dll version',version

pi=dacos(-1.0d0)

call read_inputfile(inputfile)

select case (wd%wavetype)
  case (0)	! regular airy wave
    call reg_airy_init(wd%reg,t0,t1,dt)
  case (1)	! Iregular airy wave
    call ireg_airy_init(wd%ireg,t0,t1,dt)
  case (2)	! Deterministic airy wave
    call det_airy_init(wd%det,t0,t1)
  case (3)	! Stream function
    call stream_func_init(wd%strf)
!  case (4) ! general wavemodel module
!    call wavemods_init(wd%wavemods,t0,t1,dt)
  case default
    continue
end select

end subroutine init

!-------------------------------------------------------------------------------

subroutine set_new_time(time)
use global_typesnvars
implicit none
!DEC$ ATTRIBUTES DLLEXPORT, ALIAS:'set_new_time'::set_new_time
real*8                :: time

wd%time=time

end subroutine set_new_time

!-------------------------------------------------------------------------------

subroutine get_sea_elevation(posxy_h,elevation)
use ireg_airy
use reg_airy
use det_airy
use stream_func
use global_typesnvars
use tools
use wavemods

implicit none
!DEC$ ATTRIBUTES DLLEXPORT, ALIAS:'get_sea_elevation'::get_sea_elevation
real*8,dimension(2) :: posxy_h         ! horizontal position coordinates
real*8            :: elevation		! water height above mean water level, positive upwards

real*8,dimension(3):: pos_h

pos_h(1:2)=posxy_h(1:2)
pos_h(3)=0.0d0

select case (wd%wavetype)
case (0)	! regular airy
  call reg_airy_sea_elevation(pos_h,wd%reg,elevation)
case (1)	! iregular airy
  call ireg_airy_sea_elevation(pos_h,wd%ireg,elevation)
case (2)	! deterministic irregular airy
  call det_airy_sea_elevation(pos_h,wd%det,elevation)
case (3)	! stream function
  call stream_func_sea_elevation(pos_h,wd%strf,elevation)
!case (4) ! general wavemodel module
!  call wavemods_sea_elevation(pos_h,wd%wavemods,elevation)
case default
  continue
end select

end subroutine get_sea_elevation

!-------------------------------------------------------------------------------

subroutine get_kinematics(pos_h,vel_h,acc_h,pres)
use ireg_airy
use reg_airy
use det_airy
use stream_func
use global_typesnvars
use tools
use wavemods

implicit none
!DEC$ ATTRIBUTES DLLEXPORT, ALIAS:'get_kinematics'::get_kinematics
real*8,dimension(3)            :: pos_h, &
                                  vel_h, &
								  acc_h
real*8							:: pres

select case (wd%wavetype)
case (0)	! regular airy wave
  call reg_airy_get_kinematics(wd%reg,pos_h,vel_h,acc_h,pres)
case (1)	! iregular airy wave
  call ireg_airy_get_kinematics(wd%ireg,pos_h,vel_h,acc_h,pres)
case (2)	! deterministic airy wave
  call det_airy_get_kinematics(wd%det,pos_h,vel_h,acc_h,pres)
case (3)	! stream function
  call stream_func_get_kinematics(wd%strf,pos_h,vel_h,acc_h,pres)
!case (4) ! general wavemodel module
!  call wavemods_get_kinematics(wd%wavemods,pos_h,vel_h,acc_h,pres)
case default
end select

return
10 continue
end subroutine get_kinematics


